#!/usr/bin/env python3
from PIL import Image
import sys, math
ncol=16
dither=0
args=sys.argv[1:]
fname=None
csrc=False
outname="img.seq"
def helpme(s):
    if s:
        print(s)
    print("""
Usage: %s [options] filename
Options are:
-d        - dither
-n <nr>   - limit colors (2..16)
-o <name> - output file name (default img.seq)
-c        - create colors.asm and bitmap.asm
-h        - this help""" % sys.argv[0])
    exit(0)
    
while len(args) > 0:
    arg=args.pop(0)
    if arg == '-d':
        dither=1
        continue
    if arg == "-h":
        helpme(None)
    if arg == "-o":
        if len(arg) == 0:
            helpme("Missing argument")
        outname=args.pop(0)
        continue
    if arg == '-n':
        if len(arg) == 0:
            helpme("Missing argument")
        try:
            ncol=int(args.pop(0))
        except:
            helpme("Illegal argument")
        if ncol < 2 or ncol > 16:
            helpme("Use 2 to 16 colors")
        continue
    if arg == "-c":
        csrc=True
        continue
    fname=arg
    break
if fname is None:
    helpme("Missing filename")
if args:
    helpme("Too many arguments")

cpal=[(0,0,0),(255,255,255),(136,0,0),
    (170,255,238),(204,68,204),(0,204,85),
    (0,0,170),(238,238,119),(221,136,85),(102,68,0),(255,119,119),
    (51,51,51),(119,119,119),(170,255,102),(0,136,255),(187,187,187)]
dpal=[0,0,0,255,255,255,136,0,0,170,255,238,204,68,204,0,204,85,0,0,170,238,238,119,221,136,85,102,68,0,255,119,119,51,51,51,119,119,119,170,255,102,0,136,255,187,187,187]
epal=[0,0,0,255,255,255,51,51,51,119,119,119,187,187,187]
fpal=[0,0,0,255,255,255,119,119,119,187,187,187]
image=Image.open(fname)
image=image.resize((160,200))
pal=Image.new('P',(16,16))
pal.putpalette((dpal[:3*ncol]*160)[:3*256])
im2=image.quantize(palette=pal, dither=dither)
'''
gamma 2200
tint 1000
saturation 1250
contrast 1250
brightness 1000
'''
cdif={}

def difcol(c1,c2):
    global cdif, cpal
    def sq(a):
        return a*a
    d=cdif.get((c1,c2),None)
    if d is not None:
        return d
    d=math.sqrt(
        sq(cpal[c1][0]-cpal[c2][0]) +
        sq (cpal[c1][1]-cpal[c2][1]) +
        sq(cpal[c1][2]-cpal[c2][2]))
    cdif[(c1,c2)]=d
    return d

def bests(clrs,p):
    mx=99999
    cl=0
    for n in clrs:
        d=difcol(n,p)
        if d < mx:
            mx=d
            cl=n
    return cl
    
def bestcold(x0,y0):
    global im2,ncol
    idx={}
    for x in range(4):
        for y in range(8):
            p=im2.getpixel((x0+x,y0+y)) % ncol
            if p:
                idx[p]=idx.get(p,0)+1
    idx=list(idx.items())
    idx.sort(key=lambda x:-x[1])
    if len(idx) <=3:
        return
    clrs=list(idx[i][0] for i in range(3))
    clrs.append(0)
    for x in range(4):
        for y in range(8):
            p=im2.getpixel((x0+x,y0+y)) % ncol
            if p in clrs:
                continue
            p=bests(clrs,p)
            im2.putpixel((x0+x,y0+y),p)
    #exit(0)

cidx={}
for x in range(160):
    for y in range(200):
        p=im2.getpixel((x,y)) % ncol
        cidx[p]=cidx.get(p,0)+1

cidx=list(cidx.items())
cidx.sort(key=lambda x:-x[1])
bgcolor=cidx[0][0]


def bestcol(x0,y0):
    global im2,bgcolor,ncol
    inx={}
    for x in range(4):
        for y in range(8):
            p=im2.getpixel((x0+x,y0+y)) % ncol
            if p == bgcolor:
                continue
            inx[p]=inx.get(p,0)+1
    inx=list(inx.items())
    inx.sort(key=lambda x:-x[1])
    colors=[bgcolor]
    while(len(colors) < 4 and len(inx) > 0):
        colors.append(inx.pop(0)[0])
    while(len(colors) < 4) :
        colors.append(bgcolor)
    if len(inx) == 0:
        return colors
    for x in range(4):
        for y in range(8):
            p=im2.getpixel((x0+x,y0+y)) % ncol
            if p in colors:
                continue
            p=bests(colors,p)
            im2.putpixel((x0+x,y0+y),p)
    return colors
    
colormap=[]
            
for y in range(0,200,8):
    for x in range(0,160,4):        
        color=bestcol(x,y)
        colormap.append(color)

def pixid(x,y):
    global im2,colormap,ncol
    idx = x // 4
    idy = y // 8
    cmap=colormap[idy*40+idx]
    p=im2.getpixel((x,y)) % ncol
    n=cmap.index(p)
    return n

def pixchar(x0,y0):
    px=[]
    for y in range(8):
        p=0
        for x in range(4):
            p = (p << 2) | pixid(x+x0, y+y0)
        px.append(p)
    return px

def com1(bts):
    bts=list(bts)
    cx=[]
    while(len(bts) > 0):
        c=bts.pop(0)
        n=1
        while n<255 and len(bts) > 0 and bts[0] == c:
            bts.pop(0)
            n += 1
        cx.append(n)
        cx.append(c)
    cx.append(0)
    return cx

def com2(bts):
    bts=list(bts)
    cx=[]
    while(len(bts) > 0):
        if len(bts) == 1:
            cx.append(bts.pop(0))
            cx.append(1)
            break
        if bts[0] == bts[1]:
            c=bts.pop(0)
            bts.pop(0)
            n=2
            while n<127 and len(bts) > 0 and bts[0] == c:
                bts.pop(0)
                n += 1
            cx.append(n)
            cx.append(c)
        else:
            n=1
            while n < 127 and len(bts)> n+1:
                if bts[n] == bts[n+1]:
                    break
                n = n+1
            cx.append(-n)
            cx.extend(bts[:n])
            bts = bts[n:]
    cx.append(0)
    return cx
            

def genasm():
    global colormap, bgcolor, csrc,outname
    seqf=[bgcolor]
    cmap=list(x[3] for x in colormap)
    seqf.extend(cmap)
    lines=['colormap:']
    while(len(cmap) > 0):
        cms=cmap[:8]
        #cms=[0,0,0,0,0,0,0,0]
        cmap=cmap[8:]
        byts=', '.join(list(str(x) for x in cms))
        lines.append('.byte '+byts)
    lines.append('screenmap:')
    cmap=list((x[1] << 4) | x[2] for x in colormap)
    seqf.extend(cmap)
    while(len(cmap) > 0):
        cms=cmap[:8]
        #cms=[0,0,0,0,0,0,0,0]
        cmap=cmap[8:]
        byts=', '.join(list(str(x) for x in cms))
        lines.append('.byte '+byts)
    lines.append('bgcolor .byte %d' % bgcolor)
    if csrc:
        f=open('colors.asm','w')
        w='\n'.join(lines)
        f.write(w+'\n')
        f.close()

    lines=[]
    pax=[]
    for y in range(0,200,8):
        for x in range(0,160,4):
            px=pixchar(x,y)
            pax.extend(px)
            line='.byte '+ (', '.join(str(n) for n in px))
            lines.append(line)
    if csrc:
        f=open('bitmap.asm','w')
        lines='\n'.join(lines)
        f.write(lines+'\n')
        f.close()
    seqf.extend(pax)
    bts=bytes(seqf)
    f=open(outname,'wb')
    f.write(bts)
    f.close()
    return

genasm()
im2=im2.resize((320,200))
im2.show('')

